import re
import unittest
import xeger

class TestXeger(unittest.TestCase):
    def setUp(self):
        pass

    def match(self, pattern):
        for _ in range(100):
            assert re.match(pattern, xeger.xeger(pattern))

    def test_single_dot(self):
        """
        Verify that the dot character produces only a single character.
        """
        self.match(r'^.$')

    def test_dot(self):
        """
        Verify that the dot character doesn't produce newlines.
        See: https://bitbucket.org/leapfrogdevelopment/rstr/issue/1/
        """
        for i in range(100):
            self.match(r'.+')

    def test_date(self):
        self.match(r'^([1-9]|0[1-9]|[12][0-9]|3[01])\D([1-9]|0[1-9]|1[012])\D(19[0-9][0-9]|20[0-9][0-9])$')

    def test_up_to_closing_tag(self):
        self.match(r'([^<]*)')

    def test_ipv4(self):
        self.match(r'^(\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5])\.(\d|[1-9]\d|1\d\d|2[0-4]\d|25[0-5]){3}$')

    def test_year_1900_2099(self):
        self.match(r'^(19|20)[\d]{2,2}$')

    def test_positive_or_negative_number(self):
        self.match(r'^-{0,1}\d*\.{0,1}\d+$')

    def test_positive_integers(self):
        self.match(r'^\d+$')

    def test_email_complicated(self):
        self.match(r'^([0-9a-zA-Z]([\+\-_\.][0-9a-zA-Z]+)*)+"@(([0-9a-zA-Z][-\w]*[0-9a-zA-Z]*\.)+[a-zA-Z0-9]{2,17})$')

    def test_email(self):
        self.match(r'(.*?)\@(.*?)\.(.*?)')

    def test_alpha(self):
        self.match(r'[:alpha:]')

    def test_zero_or_more_anything_non_greedy(self):
        self.match(r'(.*?)')

    def test_zero_or_more_greedy(self):
        self.match(r'(.*)')

    def test_literals(self):
        self.match(r'foo')

    def test_digit(self):
        self.match(r'\d')

    def test_nondigits(self):
        self.match(r'\D')

    def test_literal_with_repeat(self):
        self.match(r'A{3}')

    def test_literal_with_range_repeat(self):
        self.match(r'A{2, 5}')

    def test_word(self):
        self.match(r'\w')

    def test_nonword(self):
        self.match(r'\W')

    def test_or(self):
        self.match(r'foo|bar')

    def test_or_with_subpattern(self):
        self.match(r'(foo|bar)')

    def test_range(self):
        self.match(r'[A-F]')

    def test_character_group(self):
        self.match(r'[ABC]')

    def test_caret(self):
        self.match(r'^foo')

    def test_dollarsign(self):
        self.match(r'foo$')

    def test_not_literal(self):
        self.match(r'[^a]')

    def test_negation_group(self):
        self.match(r'[^AEIOU]')

    def test_lookahead(self):
        self.match(r'foo(?=bar)')

    def test_lookbehind(self):
        pattern = r'(?<=foo)bar'
        assert re.search(pattern, xeger.xeger(pattern))

    def test_backreference(self):
        self.match(r'(foo|bar)baz\1')

    def test_zero_or_more_greedy(self):
        self.match(r'a*')

    def test_zero_or_more_non_greedy(self):
        self.match(r'a*?')



"""Library to generate random strings from regular expressions.
Code borrowed and cleaned up from the python module rstr:
https://bitbucket.org/leapfrogdevelopment/rstr/

In turn inspired by the Java library Xeger:
http://code.google.com/p/xeger/
"""
import re
import sys
import string
import itertools
from random import choice, randint

if sys.version_info[0] >= 3:
    unichr = chr
    xrange = range


class Xeger(object):
    def __init__(self, limit=10):
        super(Xeger, self).__init__()
        self._limit = limit
        self._cache = dict()

        self._alphabets = {
            'printable': string.printable,
            'letters': string.ascii_letters,
            'uppercase': string.ascii_uppercase,
            'lowercase': string.ascii_lowercase,
            'digits': string.digits,
            'punctuation': string.punctuation,
            'nondigits': string.ascii_letters + string.punctuation,
            'nonletters': string.digits + string.punctuation,
            'whitespace': string.whitespace,
            'nonwhitespace': string.printable.strip(),
            'normal': string.ascii_letters + string.digits + ' ',
            'word': string.ascii_letters + string.digits + '_',
            'nonword': ''.join(set(string.printable)
                            .difference(string.ascii_letters +
                                        string.digits + '_')),
            'postalsafe': string.ascii_letters + string.digits + ' .-#/',
            'urlsafe': string.ascii_letters + string.digits + '-._~',
            'domainsafe': string.ascii_letters + string.digits + '-'
        }

        self._categories = {
            "category_digit": lambda: self._alphabets['digits'],
            "category_not_digit": lambda: self._alphabets['nondigits'],
            "category_space": lambda: self._alphabets['whitespace'],
            "category_not_space": lambda: self._alphabets['nonwhitespace'],
            "category_word": lambda: self._alphabets['word'],
            "category_not_word": lambda: self._alphabets['nonword'],
        }

        self._cases = {
            "literal": lambda x: unichr(x),
            "not_literal":
                lambda x: choice(string.printable.replace(unichr(x), '')),
            "at": lambda x: '',
            "in": lambda x: self._handle_in(x),
            "any": lambda x: choice(string.printable.replace('\n', '')),
            "range": lambda x: [unichr(i) for i in xrange(x[0], x[1] + 1)],
            "category": lambda x: self._categories[x](),
            'branch':
                lambda x: ''.join(self._handle_state(i) for i in choice(x[1])),
            "subpattern": lambda x: self._handle_group(x),
            "assert": lambda x: ''.join(self._handle_state(i) for i in x[1]),
            "assert_not": lambda x: '',
            "groupref": lambda x: self._cache[x],
            'min_repeat': lambda x: self._handle_repeat(*x),
            'max_repeat': lambda x: self._handle_repeat(*x),
            'negate': lambda x: [False],
        }

    def xeger(self, string_or_regex):
        try:
            pattern = string_or_regex.pattern
        except AttributeError:
            pattern = string_or_regex

        parsed = re.sre_parse.parse(pattern)
        result = self._build_string(parsed)
        self._cache.clear()
        return result

    def _build_string(self, parsed):
        newstr = []
        for state in parsed:
            newstr.append(self._handle_state(state))
        return ''.join(newstr)

    def _handle_state(self, state):
        opcode, value = state
        return self._cases[opcode](value)

    def _handle_group(self, value):
        result = ''.join(self._handle_state(i) for i in value[1])
        if value[0]:
            self._cache[value[0]] = result
        return result

    def _handle_in(self, value):
        candidates = list(itertools.chain(*(self._handle_state(i) for i in value)))
        if candidates[0] is False:
            candidates = set(string.printable).difference(candidates[1:])
            return choice(list(candidates))
        else:
            return choice(candidates)

    def _handle_repeat(self, start_range, end_range, value):
        result = []
        end_range = min((end_range, self._limit))
        times = randint(start_range, end_range)
        for i in xrange(times):
            result.append(''.join(self._handle_state(i) for i in value))
        return ''.join(result)
